"""
Django settings for fleet_manager project.

Generated by 'django-admin startproject' using Django 1.11.3.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os
from importlib import import_module

import env
import ldap
from django.contrib import messages
from django.urls import reverse_lazy
from django_auth_ldap.config import LDAPSearch

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env.get_secret("DJANGO_SECRET_KEY")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env.get_bool("DJANGO_DEBUG", False)

PROBES_IPS = env.get_list("DJANGO_PROBES_IP", ["0.0.0.0/0"])
ALLOWED_HOSTS = env.get_list("DJANGO_ALLOWED_HOSTS", [])
ALLOWED_HOSTS += env.get_list("DJANGO_EXTRA_ALLOWED_HOSTS", [])
DEFAULT_DOMAIN = ALLOWED_HOSTS[0] if ALLOWED_HOSTS else "localhost"

# A list of all the people who get code error notifications.
ADMINS = [(mail, mail) for mail in env.get_list("DJANGO_ADMINS", [])]

# Redirect plain HTTP requests to HTTPS.
SECURE_SSL_REDIRECT = not DEBUG

# Set this to True to avoid transmitting the CSRF cookie over HTTP accidentally.
CSRF_COOKIE_SECURE = not DEBUG

# Set this to True to avoid transmitting the session cookie over HTTP accidentally.
SESSION_COOKIE_SECURE = not DEBUG

# See https://docs.djangoproject.com/en/3.0/ref/middleware/#referrer-policy
SECURE_REFERRER_POLICY = "strict-origin-when-cross-origin"

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

DATA_UPLOAD_MAX_NUMBER_FIELDS = None


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.humanize",
    "collectfast",  # must come before staticfiles
    "django.contrib.staticfiles",
    "admin_ordering",
    "crispy_forms",
    "crispy_bootstrap5",
    "debug_toolbar",
    "django_filters",
    "django_json_widget",
    "django_q",
    "drf_spectacular",
    "drf_spectacular_sidecar",
    "netfields",
    "rest_framework",
    "social_django",
    "fleet.apps.FleetConfig",
    "powerdns.apps.PowerDnsConfig",
    "pxe.apps.PXEConfig",
    "salt.apps.SaltConfig",
    "sessions_state.apps.SessionsStateConfig",
]

MIDDLEWARE = [
    "fleet_manager.middleware.ProbesMiddleware",
    "fleet_manager.middleware.XRealIPMiddleware",
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "social_django.middleware.SocialAuthExceptionMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "fleet_manager.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "fleet_manager.wsgi.application"

LOG_LEVEL = env.get_string("DJANGO_LOG_LEVEL", "DEBUG" if DEBUG else "WARNING").upper()

# A data structure containing configuration information. The contents of this
# data structure will be passed as the argument to the configuration method
# described in LOGGING_CONFIG.

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "django.server": {
            "()": "django.utils.log.ServerFormatter",
            "format": "[{server_time}] {message}",
            "style": "{",
        },
        "default": {
            "()": "logging.Formatter",
            "format": "[{asctime}][{levelname}:{name}] {message}",
            "style": "{",
        },
    },
    "filters": {
        "require_debug_false": {"()": "django.utils.log.RequireDebugFalse"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "django.server": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "django.server",
        },
        "mail_admins": {
            "level": "ERROR",
            "filters": ["require_debug_false"],
            "class": "django.utils.log.AdminEmailHandler",
        },
    },
    "loggers": {
        "django": {
            "handlers": ["console", "mail_admins"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "django.server": {
            "handlers": ["django.server"],
            "level": "INFO",
            "propagate": False,
        },
    },
    "root": {"handlers": ["console"], "level": LOG_LEVEL},
}


# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": env.get_string("DB_NAME", "fleet_manager"),
        "USER": env.get_string("DB_USER", "fleet_manager"),
        "PASSWORD": env.get_secret("DB_PASSWORD"),
        "HOST": env.get_string("DB_HOST"),
        "PORT": env.get_int("DB_PORT", 5432),
    },
    "powerdns": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": env.get_string("DB_POWERDNS_NAME", "powerdns"),
        "USER": env.get_string("DB_POWERDNS_USER", "powerdns"),
        "PASSWORD": env.get_secret("DB_POWERDNS_PASSWORD"),
        "HOST": env.get_string("DB_POWERDNS_HOST"),
        "PORT": env.get_int("DB_POWERDNS_PORT", 5432),
    },
    "salt": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": env.get_string("DB_SALT_NAME", "salt"),
        "USER": env.get_string("DB_SALT_USER", "salt"),
        "PASSWORD": env.get_secret("DB_SALT_PASSWORD"),
        "HOST": env.get_string("DB_SALT_HOST"),
        "PORT": env.get_int("DB_SALT_PORT", 5432),
    },
}

DATABASE_ROUTERS = ("fleet_manager.database_routers.AppRouter",)


# Caches
# https://docs.djangoproject.com/en/3.0/topics/cache/

CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": env.get_string("DJANGO_CACHE_URL"),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        },
    }
}


DEFAULT_AUTO_FIELD = "django.db.models.AutoField"


# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",  # noqa: E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "Europe/Paris"

USE_I18N = True

USE_L10N = False

USE_TZ = True

DATETIME_FORMAT = "Y-m-d H:i:s"

DATE_FORMAT = "Y-m-d"


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

COLLECTFAST_DEBUG = DEBUG
COLLECTFAST_STRATEGY = "collectfast.strategies.boto3.Boto3Strategy"
STATICFILES_STORAGE = "storages.backends.s3boto3.S3Boto3Storage"
DEFAULT_FILE_STORAGE = "storages.backends.s3boto3.S3Boto3Storage"

AWS_ACCESS_KEY_ID = env.get_secret("S3_ACCESS_KEY")
AWS_SECRET_ACCESS_KEY = env.get_secret("S3_SECRET_KEY")
AWS_STORAGE_BUCKET_NAME = env.get_string("S3_BUCKET", "cri-fleet-manager")
AWS_S3_ENDPOINT_URL = env.get_string("S3_ENDPOINT")
AWS_S3_CUSTOM_DOMAIN = env.get_string("S3_CUSTOM_DOMAIN", "") or None
AWS_S3_SECURE_URLS = env.get_bool("S3_SECURE_URLS", True)
AWS_S3_URL_PROTOCOL = "https:" if AWS_S3_SECURE_URLS else "http:"

STATIC_URL = (
    ("https://" if AWS_S3_SECURE_URLS else "http://")
    + (AWS_S3_CUSTOM_DOMAIN or "localhost:8020/cri-fleet-manager")
    + "/"
)

STATICFILES_DIRS = (("node_modules", "/app/node_modules/"),)

AWS_QUERYSTRING_AUTH = True
AWS_S3_OBJECT_PARAMETERS = {"ACL": "public-read"}


# Tests

TEST_OUTPUT_DIR = os.path.join(BASE_DIR, "test-reports/")


MESSAGE_TAGS = {
    messages.DEBUG: "secondary",
    messages.INFO: "primary",
    messages.SUCCESS: "success",
    messages.WARNING: "warning",
    messages.ERROR: "danger",
}


# Auth

LOGIN_URL = reverse_lazy("social:begin", kwargs={"backend": "epita"})
LOGOUT_URL = reverse_lazy("social:disconnect", kwargs={"backend": "epita"})

LOGIN_REDIRECT_URL = "/"
LOGOUT_REDIRECT_URL = "/"
LOGIN_ERROR_URL = "/"

AUTHENTICATION_BACKENDS = (
    "social_auth_backend_epita.backend.EpitaOpenIdConnect",
    "django_auth_ldap.backend.LDAPBackend",
    "django.contrib.auth.backends.ModelBackend",
)

# OIDC

SOCIAL_AUTH_EPITA_KEY = env.get_secret("EPITA_OPENID_KEY")
SOCIAL_AUTH_EPITA_SECRET = env.get_secret("EPITA_OPENID_SECRET")

SOCIAL_AUTH_EPITA_SCOPE = ["epita", "roles", "profile", "email"]

SOCIAL_AUTH_POSTGRES_JSONFIELD = True

SOCIAL_AUTH_PIPELINE = (
    "social_core.pipeline.social_auth.social_details",
    "social_core.pipeline.social_auth.social_uid",
    "social_core.pipeline.social_auth.auth_allowed",
    "social_auth_backend_epita.pipeline.deny_old_users",
    "social_core.pipeline.social_auth.social_user",
    "social_core.pipeline.user.get_username",
    "social_auth_backend_epita.pipeline.merge_old_users",
    "social_core.pipeline.user.create_user",
    "social_core.pipeline.social_auth.associate_user",
    "social_core.pipeline.social_auth.load_extra_data",
    "social_core.pipeline.user.user_details",
    "social_auth_backend_epita.pipeline.update_email",
    "fleet_manager.pipeline.save_all_claims_as_extra_data",
    "fleet_manager.pipeline.update_groups",
    "fleet_manager.pipeline.set_permissions",
)


# LDAP
# Import and return the group type based on string name
def _import_group_type(group_type_name):
    mod = import_module("django_auth_ldap.config")
    try:
        return getattr(mod, group_type_name)()
    except:
        return None


AUTH_LDAP_SERVER_URI = env.get_string("AUTH_LDAP_SERVER_URI", "")

# The following may be needed if you are binding to Active Directory.
AUTH_LDAP_CONNECTION_OPTIONS = {ldap.OPT_REFERRALS: 0}

# Set the DN and password for the NetBox service account.
AUTH_LDAP_BIND_DN = env.get_secret("AUTH_LDAP_BIND_DN", "")
AUTH_LDAP_BIND_PASSWORD = env.get_secret("AUTH_LDAP_BIND_PASSWORD", "")

# Set a string template that describes any user’s distinguished name based on
# the username.
AUTH_LDAP_USER_DN_TEMPLATE = env.get_string("AUTH_LDAP_USER_DN_TEMPLATE", None)

# Bind with the authenticating user.
AUTH_LDAP_BIND_AS_AUTHENTICATING_USER = env.get_bool(
    "AUTH_LDAP_BIND_AS_AUTHENTICATING_USER", False
)

# Enable STARTTLS for ldap authentication.
AUTH_LDAP_START_TLS = env.get_bool("AUTH_LDAP_START_TLS", False)

# Include this setting if you want to ignore certificate errors. This might be
# needed to accept a self-signed cert.
LDAP_IGNORE_CERT_ERRORS = env.get_bool("LDAP_IGNORE_CERT_ERRORS", False)

AUTH_LDAP_USER_SEARCH_BASEDN = env.get_string("AUTH_LDAP_USER_SEARCH_BASEDN", "")
AUTH_LDAP_USER_SEARCH_ATTR = env.get_string(
    "AUTH_LDAP_USER_SEARCH_ATTR", "sAMAccountName"
)
AUTH_LDAP_USER_SEARCH = LDAPSearch(
    AUTH_LDAP_USER_SEARCH_BASEDN,
    ldap.SCOPE_SUBTREE,
    "(" + AUTH_LDAP_USER_SEARCH_ATTR + "=%(user)s)",
)

AUTH_LDAP_GROUP_ENABLE = env.get_bool("AUTH_LDAP_GROUP_ENABLE", False)

if AUTH_LDAP_GROUP_ENABLE:
    # This search ought to return all groups to which the user belongs.
    # django_auth_ldap uses this to determine group hierarchy.
    AUTH_LDAP_GROUP_SEARCH_BASEDN = env.get_string("AUTH_LDAP_GROUP_SEARCH_BASEDN", "")
    AUTH_LDAP_GROUP_SEARCH_CLASS = env.get_string(
        "AUTH_LDAP_GROUP_SEARCH_CLASS", "group"
    )
    AUTH_LDAP_GROUP_SEARCH = LDAPSearch(
        AUTH_LDAP_GROUP_SEARCH_BASEDN,
        ldap.SCOPE_SUBTREE,
        "(objectClass=" + AUTH_LDAP_GROUP_SEARCH_CLASS + ")",
    )
    AUTH_LDAP_GROUP_TYPE = _import_group_type(
        env.get_string("AUTH_LDAP_GROUP_TYPE", "GroupOfNamesType")
    )

    # Define a group required to login.
    AUTH_LDAP_REQUIRE_GROUP = env.get_string("AUTH_LDAP_REQUIRE_GROUP_DN", None)

    # Define special user types using groups. Exercise great caution when assigning
    # superuser status.
    AUTH_LDAP_USER_FLAGS_BY_GROUP = {}

    if AUTH_LDAP_REQUIRE_GROUP is not None:
        AUTH_LDAP_USER_FLAGS_BY_GROUP = {
            "is_active": environ.get("AUTH_LDAP_REQUIRE_GROUP_DN", ""),
            "is_staff": environ.get("AUTH_LDAP_IS_STAFF_DN", ""),
            "is_superuser": environ.get("AUTH_LDAP_IS_SUPERUSER_DN", ""),
        }

    # For more granular permissions, we can map LDAP groups to Django groups.
    AUTH_LDAP_FIND_GROUP_PERMS = env.get_bool("AUTH_LDAP_FIND_GROUP_PERMS", True)
    AUTH_LDAP_MIRROR_GROUPS = env.get_bool("AUTH_LDAP_MIRROR_GROUPS", None)

# Cache groups for one hour to reduce LDAP traffic
AUTH_LDAP_CACHE_TIMEOUT = env.get_int("AUTH_LDAP_CACHE_TIMEOUT", 3600)

# Populate the Django user from the LDAP directory.
AUTH_LDAP_USER_ATTR_MAP = {
    "first_name": env.get_string("AUTH_LDAP_ATTR_FIRSTNAME", "givenName"),
    "last_name": env.get_string("AUTH_LDAP_ATTR_LASTNAME", "sn"),
    "email": env.get_string("AUTH_LDAP_ATTR_MAIL", "mail"),
}


# Django debug toolbar


def show_toolbar(request):
    from django.conf import settings  # pylint:disable=import-outside-toplevel

    return settings.DEBUG


DEBUG_TOOLBAR_CONFIG = {
    "SHOW_TOOLBAR_CALLBACK": "fleet_manager.settings.show_toolbar",
    "SHOW_COLLAPSED": True,
}

# Email settings

DEFAULT_FROM_EMAIL = env.get_string(
    "DJANGO_DEFAULT_FROM_EMAIL", f"noreply@{DEFAULT_DOMAIN}"
)

if env.get_bool("DJANGO_DEBUG", False):
    EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"
else:
    EMAIL_HOST = env.get_string("DJANGO_SMTP_HOSTNAME", "localhost")
    EMAIL_PORT = env.get_string("DJANGO_SMTP_PORT", 25)
    EMAIL_HOST_USER = env.get_string("DJANGO_SMTP_USER", "")
    EMAIL_HOST_PASSWORD = env.get_secret("DJANGO_SMTP_PASSWORD", "")
    EMAIL_USE_TLS = env.get_bool("DJANGO_SMTP_STARTTLS", False)

SERVER_EMAIL = env.get_string("DJANGO_SERVER_EMAIL", f"fleet@{DEFAULT_DOMAIN}")
EMAIL_SUBJECT_PREFIX = env.get_string("DJANGO_EMAIL_SUBJECT_PREFIX", "[Django] ")


# DRF

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework_simplejwt.authentication.JWTAuthentication",
        "rest_framework.authentication.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.BasicAuthentication",
    ),
    "DEFAULT_THROTTLE_CLASSES": (
        "rest_framework.throttling.AnonRateThrottle",
        "rest_framework.throttling.UserRateThrottle",
    ),
    "DEFAULT_THROTTLE_RATES": {"anon": "100/hour", "user": "1000/hour"},
    "DEFAULT_FILTER_BACKENDS": ("fleet_manager.drf.DjangoFilterBackend",),
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
    "PAGE_SIZE": 10000,
}

SPECTACULAR_SETTINGS = {
    "title": "Fleet Manager",
    "description": "Fleet Manager API documentation",
    "VERSION": "0.1.0",
    "SERVE_PUBLIC": True,
    # offline support
    "SWAGGER_UI_DIST": "SIDECAR",
    "SWAGGER_UI_FAVICON_HREF": "SIDECAR",
    "REDOC_DIST": "SIDECAR",
}


# Django Q

Q_CLUSTER = {
    "name": env.get_string("DJANGO_Q_CLUSTER_NAME", "cri_fleet_manager"),
    "workers": env.get_int("DJANGO_Q_THREADS", 2),
    "timeout": env.get_int("DJANGO_Q_TIMEOUT", 600),
    "retry": env.get_int("DJANGO_Q_RETRY", 660),
    "django_redis": "default",
}


# PXE specific settings

DEFAULT_KERNEL_VALUE = "https://ipxe.pie.cri.epita.fr/{name}_bzImage"
DEFAULT_INITRD_VALUE = "https://ipxe.pie.cri.epita.fr/{name}_initrd"
DEFAULT_CMDLINE_VALUE = "vga=0x31a loglevel=4 initrd={name}_initrd"
CMDLINE_OVERRIDE = "net.ifnames=0"

# Sessions state specific settings
SESSIONS_MERGE_TIMEOUT_SECONDS = env.get_int("SESSIONS_MERGE_TIMEOUT_SECONDS", 3600)
SESSION_LOG_MAX_AGE = 365


# Kea specific settings

KEA_SHARED_SECRET = env.get_secret("KEA_SHARED_SECRET")
# These are used to generate Kea's config
DHCP_VALID_LIFETIME = env.get_int("DHCP_VALID_LIFETIME", 3600)
DHCP_MIN_VALID_LIFETIME = env.get_int("DHCP_MIN_VALID_LIFETIME", 360)
DHCP_MAX_VALID_LIFETIME = env.get_int("DHCP_MAX_VALID_LIFETIME", 7200)
DHCP_CONTROL_SOCKET_PATH = env.get_string(
    "DHCP_CONTROL_SOCKET_PATH", "/shared/kea4-ctrl-socket"
)
DHCP_LOG_LEVEL = env.get_string("DHCP_LOG_LEVEL", "INFO")
DHCP_DEBUG_LEVEL = env.get_int("DHCP_DEBUG_LEVEL", 0)

DHCP_DOMAIN_NAME_SERVERS = env.get_list(
    "DHCP_DOMAIN_NAME_SERVERS", ["1.1.1.1", "1.0.0.1"]
)
DHCP_NTP_SERVERS = env.get_list("DHCP_NTP_SERVERS", [])
DHCP_NEXT_SERVER = env.get_string("DHCP_NEXT_SERVER")


# DNS specific settings

DNS_SOA_MNAME = env.get_string("DNS_SOA_MNAME", "a.misconfigured.dns.server.invalid")
DNS_SOA_RNAME = env.get_string("DNS_SOA_RNAME", None)
DNS_SOA_REFRESH = env.get_int("DNS_SOA_REFRESH", 14400)
DNS_SOA_RETRY = env.get_int("DNS_SOA_RETRY", 3600)
DNS_SOA_EXPIRE = env.get_int("DNS_SOA_EXPIRE", 604800)
DNS_SOA_TTL = env.get_int("DNS_SOA_TTL", 300)

DNS_NS_NAME = env.get_string("DNS_NS_NAME", "a.misconfigured.dns.server.invalid")

DNS_RECORD_DEFAULT_TTL = env.get_int("DNS_RECORD_DEFAULT_TTL", DNS_SOA_TTL)


# Salt specific settings

SALT_BACKEND_CONFIG = {
    "endpoint": env.get_string("SALT_API_ENDPOINT"),
    "sharedsecret": env.get_secret("SALT_API_SHARED_SECRET"),
    "ignore_ssl_errors": env.get_bool("SALT_API_IGNORE_SSL_ERRORS", True),
}

CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap5"
CRISPY_TEMPLATE_PACK = "bootstrap5"
